/*
 * standalone quality check for given password
 * 
 * pwdqc.c
 * ww@mpi-inf.mpg.de
 */

#define VERSION "1.2"

#include <unistd.h>
#include <limits.h>
#include <bits/posix1_lim.h>
#include <bits/local_lim.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <libgen.h>
#include "passwdqc.h"



void
usage(const char *name)
{
  char * shortname = basename((char *)name);
  if (shortname == NULL) {
    shortname=(char *)name;
  }
  fprintf(stderr,"usage: %s [-h|-?] [ -u <username>] [-o] \n\t\t[min=<number|disabled>[,<number>|disabled]...] \n\t\t[max=<number>] [passphrase=<number of passphrase words>]\n\t\t[match=<number>] [similar=permit|deny] [random=<number>[only]]\n", shortname);
  fprintf(stderr,"\tIf -o is present, the old password is expected as first line on stdin.\n");
}

/* remove possible trailing newline */
void
strip_nl(char * strng)
{
  size_t last_index=0;

  last_index=strlen(strng) - 1;
  if ( strng[last_index] == '\n') {
    strng[last_index] = '\0';
  }
}

int
main (int argc, char ** argv)
{
  passwdqc_params_t params;
  char * reason;
  struct passwd *pw = NULL;
  char * oldpass = NULL;
  char * newpass = NULL;
  long pwbuflen = 0l;
  int argnum = 0;
  const char *p;
  char *e;
  int i;
  unsigned long v;
  int oldpass_check = 0;



  /* Set parameter defaults */
  params.max              = 40;
  params.min[0]           = params.max + 5; /* a.k.a. disabled */
  params.min[1]           = 14;
  params.min[2]           = 12;
  params.min[3]           =  8;
  params.min[4]           =  8;
  params.passphrase_words =  3;
  params.match_length     =  4;
  params.similar_deny     =  1;
  params.random_bits      = 42;

  /* After first defaults, collect command line arguments*/

  for (argnum = 1; argnum < argc ; argnum++){
    /* User name for passwd information */
    if (!strncmp(argv[argnum], "-u",2)) {
      argnum++;
      if (argnum == argc){
	fprintf(stderr, "-u flag seen without corresponding user argument\n");
	usage(argv[0]);
	exit(1);
      }
      errno=0;
      pw=getpwnam(argv[argnum]);
      if (pw == NULL) {
	fprintf(stderr,"Could not find passwd entry for user %s\n",argv[argnum]);
	perror("getpwnam");
	exit(1);
      }
      /* Old password for comparison */

    } else if (!strncmp(argv[argnum], "-o",2)) {
      oldpass_check = 1;
      /* Version */
    } else if (!strncmp(argv[argnum], "-v",2) \
	       || !strncmp(argv[argnum], "--version", 9)) {
      fprintf(stderr, "Version:%s\n", VERSION);
      exit(0);
      /* Help */
    } else if (!strncmp(argv[argnum], "-h",2) \
       || !strncmp(argv[argnum], "--help",6) \
       || !strncmp(argv[argnum], "-?",2)) {
      usage(argv[0]);
      exit(0);
      /* Unknown option */
    } else if (!strncmp(argv[argnum], "-",1)) {
      fprintf(stderr,"Unknown flag %s\n",argv[argnum]);
      usage(argv[0]);
      exit(1);
      /* This is stolen from pam_passwdqc.c */
      /* Set params for check from command line */
    } else if (!strncmp(argv[argnum], "min=", 4)) {
      p = &argv[argnum][4];
      for (i = 0; i < 5; i++) {
	if (!strncmp(p, "disabled", 8)) {
	  v = INT_MAX;
	  p += 8;
	} else {
	  v = strtoul(p, &e, 10);
	  p = e;
	}
	if (i < 4 && *p++ != ',') break;
	if (v > INT_MAX) {
	  fprintf(stderr, "Ignoring min parameter %lu, more than maximum %ul\n",v,INT_MAX);
	  break;
	}
	if (i && (int)v > params.min[i - 1]) {
	  fprintf(stderr, "Ignoring min parameter %lu, not smaller than %ul\n",v,params.min[i - 1]);
	  break;
	}
	params.min[i] = v;
      }
      if (*p) {
	fprintf(stderr,"Excessive arguments to min=\n");
	usage(argv[0]);
	exit(1);
      }
    } else if (!strncmp(argv[argnum], "max=", 4)) {
      v = strtoul(argv[argnum] + 4, &e, 10);
      if (*e || v < 8 || v > INT_MAX) {
	fprintf(stderr,"Argument to max= out of range (8<arg<%d)\n",INT_MAX);
	usage(argv[0]);
	exit(1);	
      }
      params.max = v;
    } else if (!strncmp(argv[argnum], "passphrase=", 11)) {
      v = strtoul(argv[argnum] + 11, &e, 10);
      if (*e || v > INT_MAX) {
	fprintf(stderr,"Argument to passphrase= out of range (arg<%d)\n",INT_MAX);
	usage(argv[0]);
	exit(1);	
      }
      params.passphrase_words = v;
    } else if (!strncmp(argv[argnum], "match=", 6)) {
      v = strtoul(argv[argnum] + 6, &e, 10);
      if (*e || v > INT_MAX){
	fprintf(stderr,"Argument to match= out of range (arg<%d)\n",INT_MAX);
	usage(argv[0]);
	exit(1);	
      }
      params.match_length = v;
    } else if (!strncmp(argv[argnum], "similar=", 8)) {
      if (!strcmp(argv[argnum] + 8, "permit"))
	params.similar_deny = 0;
      else if (!strcmp(argv[argnum] + 8, "deny"))
	params.similar_deny = 1;
      else {
	fprintf(stderr,"Argument to similar= must be \"permit\" or \"deny\" \n");
	usage(argv[0]);
	exit(1);	
      }
    } else if (!strncmp(argv[argnum], "random=", 7)) {
      v = strtoul(argv[argnum] + 7, &e, 10);
      if (!strcmp(e, ",only")) {
	e += 5;
	params.min[4] = INT_MAX;
      }
      if (*e || v > INT_MAX){
	fprintf(stderr,"Argument to random= out of range (arg<%d)\n",INT_MAX);
	usage(argv[0]);
	exit(1);	
      }
      params.random_bits = v;
    } 
  }
  
  pwbuflen=sysconf(_SC_LOGIN_NAME_MAX);


  if ( (long)(int)pwbuflen != pwbuflen) {
    fprintf (stderr, "Overflow of system maximal password langth %ld (int version != long version).\n", pwbuflen);
    exit(1);
  }

  if ( (int)pwbuflen <= _POSIX_LOGIN_NAME_MAX) {
    fprintf (stderr, "Unreasonable short system maximal password langth %ld.\n", pwbuflen);
    exit(1);
  }

  if (oldpass_check){
    if ((oldpass = malloc(pwbuflen)) == NULL){
      perror("malloc");
      exit (1);
    }
    
    if (fgets(oldpass, (int)pwbuflen, stdin)==NULL) {
      fprintf (stderr, "Could not get old password to check against.\n");
      exit(1);    
    }
    strip_nl(oldpass);
  }

  if ((newpass = malloc(pwbuflen)) == NULL){
    perror("malloc");
    exit (1);
  }

  if (fgets(newpass, (int)pwbuflen, stdin)==NULL) {
    fprintf (stderr, "Could not get any password to check.\n");
    exit(1);    
  }
  strip_nl(newpass);

  reason=_passwdqc_check(& params, newpass, oldpass, pw);

  if (reason) {
    printf("%s\n", reason);
    exit(1);
  }
  exit(0);
}
